Legend:
Page
Library
Module
Module type
Parameter
Class
Class type
Page
Library
Module
Module type
Parameter
Class
Class type
send_all and send_msg_all now pass their messages correctly and use the sender queue. Add multipart send/receive coverage (#132, @ComanderP).Handle EINTR without leaving sockets in an inconsistent multipart state:
Context.terminate waits for all sockets to close.Async_kernel.Ivar.fill.const char* when writing to the inside of an allocated string value (#116, @Leonidas-from-XIV)STREAM sockets (#114, @c-cube)mask_in, mask_out, mask_in_outXpub typo (#107, @roddyyaga)zmq and zmq-lwt (#101, #102)EINVAL when attempting to get the curve encryption keys (#85, #86, #91)configurator anymore (#96)makedev issue. This requires us to drop support for OCaml 4.03 (#93)Msg.gets (#87, #90)/usr/localZMQ.Context.terminateget_bytes_optionget_identity/set_identity socket-type restrictions